#!/usr/bin/env python
# filename: germline_dbs.py

#
# Copyright (c) 2016 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from argparse import ArgumentParser
import os
import platform
import shutil
import subprocess as sp
import sys

from Bio import SeqIO


def parse_arguments():
    parser = ArgumentParser(
        "Creates AbStar germline databases using an IMGT-gapped FASTA files of germline sequences. \
        Properly formatted germline sequence files can be obtained from: http://www.imgt.org/genedb/"
    )
    parser.add_argument(
        "-v",
        "--variable",
        dest="v",
        required=True,
        help="Path to an IMGT-gapped, FASTA-formatted file containing Variable gene sequences. \
                        Sequences for both heavy and light chains should be included in a single file. \
                        Required.",
    )
    parser.add_argument(
        "-d",
        "--diversity",
        dest="d",
        required=True,
        help="Path to an IMGT-gapped, FASTA-formatted file containing Diversity gene sequences. \
                        Required.",
    )
    parser.add_argument(
        "-j",
        "--joining",
        dest="j",
        required=True,
        help="Path to an IMGT-gapped, FASTA-formatted file containing Joining gene sequences. \
                        Sequences for both heavy and light chains should be included in a single file. \
                        Required.",
    )
    parser.add_argument(
        "-r",
        "--receptor",
        dest="receptor",
        required=True,
        type=str.lower,
        choices=["bcr", "tcr"],
        help="Receptor type for the germline database. Choices include 'bcr' and 'tcr'. \
                        Required.",
    )
    parser.add_argument(
        "-i",
        "--isotypes",
        dest="isotypes",
        default=None,
        help="Path to a FASTA-formatted file containing isotype sequences. \
                        The name of the isotype in the FASTA file is what will be reported by AbStar, \
                        so the use of standard nomenclature (IgG, IgG1, etc) is encouraged. \
                        If an isotype file is not provided, isotypes will not be parsed by AbStar when \
                        using the germline database.",
    )
    parser.add_argument(
        "-m",
        "--manifest",
        dest="manifest",
        default=None,
        help="Path to a plain-text file containing information about the germline database. \
                        Format is not important, but this is the place for optional inforamtion like the origin \
                        of the germline database, the date of download, etc. Optional. If provided, the contents \
                        of the file will be stored as 'manifest.txt' in the top level of the germline database \
                        (for user-provided databases, that would be '~/.abstar/germline_dbs/<species>/).",
    )
    parser.add_argument(
        "-n",
        "--name",
        dest="name",
        required=True,
        help="Name of the resulting germline database. \
                        If an abstar germline database with the same already exists, it will be replaced. \
                        Germline database names are converted to lowercase, so 'Human' and 'human' are equivalent. \
                        User-added germline databases will persist even after abstar updates, so if you have added a \
                        'human' database and a new version of AbStar contains an updated 'human' database, the user-added \
                        database will still be used after the update.",
    )
    parser.add_argument(
        "-l",
        "--location",
        dest="db_location",
        default=None,
        help="Location into which the new germline databases will be deposited. \
                        Default is '~/.abstar/'. \
                        Note that AbStar will only use user-generated databases found in '~/abstar/', \
                        so this option is provided primarily to test database creation without overwriting \
                        current databases of the same name. \
                        If the directory does not exist, it will be created.",
    )
    parser.add_argument(
        "--include-species-in-gene-name",
        default=True,
        help="If `True`, the species is parsed from the IMGT header and appended to the \
                        germline gene name. If `False`, the species is not parsed, and the species name reported \
                        by abstar will be the database name ('human', for example). Default is `True`.",
    )
    parser.add_argument(
        "--allow-partials",
        dest="allow_partials",
        default=None,
        choices=("3'", "5'", "both"),
        help="If set, will include 'partial' sequences (partial in 3', partial in 5' or both) to the \
                        abstar germline database. 'Partial' is an annotation (field 14 in the IMGT/GENE-DB header) \
                        that indicates a germline sequence is incomplete, even if it is otherwise considered \
                        functional. Default is to exclude all partial sequences.",
    )
    parser.add_argument(
        "--allow-orphons",
        dest="allow_orphons",
        default=False,
        help="If set, will include 'orphon' sequences. Orphons are germline genes that are found \
                        outside traditional Ig loci and on different chromosomes. The are identified by IMGT with \
                        an '/OR' annotation in the gene name. Default is to exclude all orphons.",
    )
    parser.add_argument(
        "--allow-unlocalized",
        dest="allow_unlocalized",
        default=False,
        help="If set, will include unlocalized genes. Unlocalized genes are are identified by \
                        IMGT with the presence of a 'D' or 'S' annotation appended to the gene number. \
                        Default is to exclude all unlocalized genes.",
    )
    parser.add_argument(
        "-D",
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="More verbose logging if set.",
    )
    args = parser.parse_args()
    return args


# -------------------------
#
#    FILES/DIRECTORIES
#
# -------------------------


def get_addon_directory(db_location, receptor):
    if db_location is not None:
        print("\n")
        print(
            "NOTE: You have selected a non-default location for the germline directory."
        )
        string = "AbStar only looks in the default location (~/.abstar/) for user-created germline databases, "
        string += "so this database will not be used by AbStar. The custom database location option is primarily "
        string += "provided so that users can test the database creation process without overwriting existing databases."
        print(string)
        addon_dir = db_location
    else:
        addon_dir = os.path.expanduser("~/.abstar/germline_dbs")
    addon_dir = os.path.join(addon_dir, receptor.lower())
    if not os.path.isdir(addon_dir):
        os.makedirs(addon_dir)
    return addon_dir


def get_binary_directory():
    mod_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    bin_dir = os.path.join(mod_dir, "assigners/bin")
    return bin_dir


def check_for_existing_db(addon_dir, dbname):
    dbs = [os.path.basename(d[0]) for d in os.walk(addon_dir)]
    if dbname.lower() in dbs:
        print("\n")
        print("WARNING: A {} germline database already exists.".format(dbname.lower()))
        print("Creating a new database with that name will overwrite the old one.")
        keep_going = raw_input("Do you want to continue? [y/N]: ")
        if keep_going.upper() not in ["Y", "YES"]:
            print("")
            print("Aborting germline database creation.")
            print("\n")
            sys.exit()


def make_db_directories(addon_dir, dbname, isotypes):
    dbname_dir = os.path.join(addon_dir, dbname.lower())
    if not os.path.isdir(dbname_dir):
        os.makedirs(dbname_dir)
    db_names = ["imgt_gapped", "ungapped", "blast"]
    if isotypes is not None:
        db_names.append("isotypes")
    for db_name in db_names:
        db_dir = os.path.join(dbname_dir, db_name)
        if os.path.isdir(db_dir):
            shutil.rmtree(db_dir)
        os.makedirs(db_dir)


# -------------------------
#
#    DATABASE CREATION
#
# -------------------------


def make_blast_db(ungapped_germline_file, addon_directory, segment, dbname):
    print("  - BLASTn")
    bin_dir = get_binary_directory()
    mbd_binary = os.path.join(
        bin_dir, "makeblastdb_{}".format(platform.system().lower())
    )
    mbd_output = os.path.join(
        addon_directory, "{}/blast/{}".format(dbname.lower(), segment.lower())
    )
    mbd_log = os.path.join(
        addon_directory, "{}/blast/{}.blastlog".format(dbname.lower(), segment.lower())
    )
    mbd_cmd = "{} -in {} -out {} -parse_seqids -dbtype nucl -title {} -logfile {}".format(
        mbd_binary, ungapped_germline_file, mbd_output, segment.lower(), mbd_log
    )
    p = sp.Popen(mbd_cmd, stdout=sp.PIPE, stderr=sp.PIPE, shell=True)
    stdout, stderr = p.communicate()
    return mbd_output, stdout, stderr


def make_ungapped_db(
    ungapped_germline_file,
    addon_directory,
    segment,
    dbname,
    include_species_in_gene_name,
):
    print("  - ungapped FASTA")
    output_file = os.path.join(
        addon_directory, "{}/ungapped/{}.fasta".format(dbname.lower(), segment.lower())
    )
    seqs = SeqIO.parse(open(ungapped_germline_file), "fasta")
    if include_species_in_gene_name:
        fastas = []
        for s in seqs:
            gene = s.description.split("|")[1]
            species = s.description.split("|")[2].lower().replace(" ", "-")
            fastas.append(
                ">{}__{}\n{}".format(gene, species, str(s.seq).replace(".", ""))
            )
    else:
        fastas = [
            ">{}\n{}".format(s.description.split("|")[1], str(s.seq).replace(".", ""))
            for s in seqs
        ]
    open(output_file, "w").write("\n".join(fastas))
    return output_file


def make_imgt_gapped_db(
    input_file,
    addon_directory,
    segment,
    dbname,
    allow_partials,
    allow_orphons,
    allow_unlocalized,
):
    print("  - IMGT-gapped FASTA")
    output_file = os.path.join(
        addon_directory,
        "{}/imgt_gapped/{}.fasta".format(dbname.lower(), segment.lower()),
    )
    seqs = sorted(list(SeqIO.parse(open(input_file), "fasta")), key=lambda x: x.id)
    fastas = []
    for s in seqs:
        partial = s.description.split("|")[13].strip()
        if "3" in partial:
            if not any([allow_partials == "3'", allow_partials == "both"]):
                continue
                # fastas.append(">{}\n{}".format(s.description, str(s.seq).upper()))
        if "5" in partial:
            if not any([allow_partials == "5'", allow_partials == "both"]):
                continue
                # fastas.append(">{}\n{}".format(s.description, str(s.seq).upper()))
        if "/OR" in s.description.split("|")[1]:
            if not allow_orphons:
                continue
        gene_fam = s.description.split("|")[1].split("-")[0]
        if gene_fam[-1] in ["D", "S"]:
            if not allow_unlocalized:
                continue
        fastas.append(">{}\n{}".format(s.description, str(s.seq).upper()))
    open(output_file, "w").write("\n".join(fastas))
    return output_file


def make_isotype_db(input_file, addon_directory, dbname):
    print("  - isotypes")
    output_file = os.path.join(
        addon_directory, "{}/isotypes/isotypes.fasta".format(dbname.lower())
    )
    seqs = sorted(list(SeqIO.parse(open(input_file), "fasta")), key=lambda x: x.id)
    fastas = [">{}\n{}".format(s.id, str(s.seq).upper()) for s in seqs]
    open(output_file, "w").write("\n".join(fastas))
    return output_file


def transfer_manifest_data(manifest, addon_directory, dbname):
    output_file = os.path.join(
        addon_directory, "{}/manifest.txt".format(dbname.lower())
    )
    manifest_data = open(manifest, "r").read()
    open(output_file, "w").write(manifest_data)
    return output_file


# -------------------------
#
#        PRINTING
#
# -------------------------


def print_segment_info(segment, input_file):
    seqs = list(SeqIO.parse(open(input_file), "fasta"))
    print("\n")
    seg_string = "  " + segment.upper() + "  "
    print("-" * len(seg_string))
    print(seg_string)
    print("-" * len(seg_string))
    print(input_file)
    print("input file contains {} sequences".format(len(seqs)))
    print("")
    print("Building germline databases:")


def print_manifest_info(manifest):
    seg_string = "  MANIFEST  "
    print("\n")
    print("-" * len(seg_string))
    print(seg_string)
    print("-" * len(seg_string))
    print(manifest)
    print("")
    print("Transferring manifest data...")


# -------------------------
#
#          MAIN
#
# -------------------------


def main():
    args = parse_arguments()
    addon_dir = get_addon_directory(args.db_location, args.receptor)
    check_for_existing_db(addon_dir, args.name)
    make_db_directories(addon_dir, args.name, args.isotypes)
    for segment, infile in [
        ("Variable", args.v),
        ("Diversity", args.d),
        ("Joining", args.j),
    ]:
        print_segment_info(segment, infile)
        imgt_gapped_file = make_imgt_gapped_db(
            infile,
            addon_dir,
            segment[0].lower(),
            args.name,
            args.allow_partials,
            args.allow_orphons,
            args.allow_unlocalized,
        )
        ungapped_file = make_ungapped_db(
            imgt_gapped_file,
            addon_dir,
            segment[0].lower(),
            args.name,
            args.include_species_in_gene_name,
        )
        blast_file, stdout, stderr = make_blast_db(
            ungapped_file, addon_dir, segment[0].lower(), args.name
        )
        if args.debug:
            print(stdout)
            print(stderr)
    if args.isotypes is not None:
        print_segment_info("ISOTYPES", args.isotypes)
        isotype_file = make_isotype_db(args.isotypes, addon_dir, args.name)
    if args.manifest is not None:
        print_manifest_info(args.manifest)
        transfer_manifest_data(args.manifest, addon_dir, args.name)
    print("\n")


if __name__ == "__main__":
    main()



import warnings
from Bio import BiopythonWarning

warnings.simplefilter("ignore", BiopythonWarning)

from .core.abstar import run, run_standalone, main, create_parser, validate_args
from .utils.preprocess import fastqc, adapter_trim, quality_trim

from .version import __version__



import pyarrow as pa


# Add pyarrow schema to use when writing JSON-formatted abstar output to parquet files.
# Adding this explicitly defined schema attributes allows for nested data structures to be queried more efficiently.
schema = pa.schema(
    [
        pa.field("seq_id", pa.string()),
        pa.field("chain", pa.string()),
        pa.field(
            "v_gene",
            pa.struct(
                [
                    pa.field("full", pa.string()),
                    pa.field("fam", pa.string()),
                    pa.field("gene", pa.string()),
                    pa.field("score", pa.int64()),
                    pa.field("assigner_score", pa.float64()),
                    pa.field(
                        "others",
                        pa.list_(
                            pa.struct(
                                [
                                    pa.field("full", pa.string()),
                                    pa.field("assigner_score", pa.float64()),
                                ]
                            )
                        ),
                    ),
                ]
            ),
        ),
        pa.field(
            "d_gene",
            pa.struct(
                [
                    pa.field("full", pa.string()),
                    pa.field("fam", pa.string()),
                    pa.field("gene", pa.string()),
                    pa.field("score", pa.int64()),
                    pa.field("assigner_score", pa.int64()),
                    pa.field(
                        "others",
                        pa.list_(
                            pa.struct(
                                [
                                    pa.field("full", pa.string()),
                                    pa.field("assigner_score", pa.int64()),
                                ]
                            )
                        ),
                    ),
                ]
            ),
        ),
        pa.field(
            "j_gene",
            pa.struct(
                [
                    pa.field("full", pa.string()),
                    pa.field("gene", pa.string()),
                    pa.field("score", pa.int64()),
                    pa.field("assigner_score", pa.float64()),
                    pa.field(
                        "others",
                        pa.list_(
                            pa.struct(
                                [
                                    pa.field("full", pa.string()),
                                    pa.field("assigner_score", pa.float64()),
                                ]
                            )
                        ),
                    ),
                ]
            ),
        ),
        pa.field(
            "assigner_scores",
            pa.struct(
                [
                    pa.field("v", pa.float64()),
                    pa.field("d", pa.int64()),
                    pa.field("j", pa.float64()),
                ]
            ),
        ),
        pa.field("vdj_assigner", pa.string()),
        pa.field("isotype", pa.string()),
        pa.field("isotype_score", pa.int64()),
        pa.field(
            "isotype_alignment",
            pa.struct(
                [
                    pa.field("query", pa.string()),
                    pa.field("midline", pa.string()),
                    pa.field("isotype", pa.string()),
                ]
            ),
        ),
        pa.field(
            "nt_identity",
            pa.struct(
                [
                    pa.field("v", pa.float64()),
                    pa.field("j", pa.float64()),
                ]
            ),
        ),
        pa.field(
            "aa_identity",
            pa.struct(
                [
                    pa.field("v", pa.float64()),
                    pa.field("j", pa.float64()),
                ]
            ),
        ),
        pa.field("junc_len", pa.int64()),
        pa.field("cdr3_len", pa.int64()),
        pa.field("vdj_nt", pa.string()),
        pa.field("gapped_vdj_nt", pa.string()),
        pa.field("fr1_nt", pa.string()),
        pa.field("cdr1_nt", pa.string()),
        pa.field("fr2_nt", pa.string()),
        pa.field("cdr2_nt", pa.string()),
        pa.field("fr3_nt", pa.string()),
        pa.field("cdr3_nt", pa.string()),
        pa.field("fr4_nt", pa.string()),
        pa.field("vdj_germ_nt", pa.string()),
        pa.field("gapped_vdj_germ_nt", pa.string()),
        pa.field("junc_nt", pa.string()),
        pa.field(
            "region_len_nt",
            pa.struct(
                [
                    pa.field("fr1", pa.int64()),
                    pa.field("cdr1", pa.int64()),
                    pa.field("fr2", pa.int64()),
                    pa.field("cdr2", pa.int64()),
                    pa.field("fr3", pa.int64()),
                    pa.field("cdr3", pa.int64()),
                    pa.field("fr4", pa.int64()),
                ]
            ),
        ),
        pa.field(
            "var_muts_nt",
            pa.struct(
                [
                    pa.field("num", pa.int64()),
                    pa.field(
                        "muts",
                        pa.list_(
                            pa.struct(
                                [
                                    pa.field("was", pa.string()),
                                    pa.field("is", pa.string()),
                                    pa.field("raw_position", pa.string()),
                                    pa.field("position", pa.string()),
                                    pa.field("codon", pa.string()),
                                ]
                            )
                        ),
                    ),
                ]
            ),
        ),
        pa.field(
            "join_muts_nt",
            pa.struct(
                [
                    pa.field("num", pa.int64()),
                    pa.field(
                        "muts",
                        pa.list_(
                            pa.struct(
                                [
                                    pa.field("was", pa.string()),
                                    pa.field("is", pa.string()),
                                    pa.field("raw_position", pa.string()),
                                    pa.field("position", pa.string()),
                                    pa.field("codon", pa.string()),
                                ]
                            )
                        ),
                    ),
                ]
            ),
        ),
        pa.field("mut_count_nt", pa.int64()),
        pa.field("vdj_aa", pa.string()),
        pa.field("fr1_aa", pa.string()),
        pa.field("cdr1_aa", pa.string()),
        pa.field("fr2_aa", pa.string()),
        pa.field("cdr2_aa", pa.string()),
        pa.field("fr3_aa", pa.string()),
        pa.field("cdr3_aa", pa.string()),
        pa.field("fr4_aa", pa.string()),
        pa.field("vdj_germ_aa", pa.string()),
        pa.field("junc_aa", pa.string()),
        pa.field(
            "region_len_aa",
            pa.struct(
                [
                    pa.field("fr1", pa.int64()),
                    pa.field("cdr1", pa.int64()),
                    pa.field("fr2", pa.int64()),
                    pa.field("cdr2", pa.int64()),
                    pa.field("fr3", pa.int64()),
                    pa.field("cdr3", pa.int64()),
                    pa.field("fr4", pa.int64()),
                ]
            ),
        ),
        pa.field(
            "var_muts_aa",
            pa.struct(
                [
                    pa.field("num", pa.int64()),
                    pa.field(
                        "muts",
                        pa.list_(
                            pa.struct(
                                [
                                    pa.field("was", pa.string()),
                                    pa.field("is", pa.string()),
                                    pa.field("raw_position", pa.string()),
                                    pa.field("position", pa.string()),
                                    pa.field("codon", pa.string()),
                                ]
                            )
                        ),
                    ),
                ]
            ),
        ),
        pa.field(
            "join_muts_aa",
            pa.struct(
                [
                    pa.field("num", pa.int64()),
                    pa.field(
                        "muts",
                        pa.list_(
                            pa.struct(
                                [
                                    pa.field("was", pa.string()),
                                    pa.field("is", pa.string()),
                                    pa.field("raw_position", pa.string()),
                                    pa.field("position", pa.string()),
                                    pa.field("codon", pa.string()),
                                ]
                            )
                        ),
                    ),
                ]
            ),
        ),
        pa.field(
            "region_muts_nt",
            pa.struct(
                [
                    pa.field(
                        "fr1",
                        pa.struct(
                            [
                                pa.field("num", pa.int64()),
                                pa.field(
                                    "muts",
                                    pa.list_(
                                        pa.struct(
                                            [
                                                pa.field("was", pa.string()),
                                                pa.field("is", pa.string()),
                                                pa.field("raw_position", pa.string()),
                                                pa.field("position", pa.string()),
                                                pa.field("codon", pa.string()),
                                            ]
                                        )
                                    ),
                                ),
                            ]
                        ),
                    ),
                    pa.field(
                        "cdr1",
                        pa.struct(
                            [
                                pa.field("num", pa.int64()),
                                pa.field(
                                    "muts",
                                    pa.list_(
                                        pa.struct(
                                            [
                                                pa.field("was", pa.string()),
                                                pa.field("is", pa.string()),
                                                pa.field("raw_position", pa.string()),
                                                pa.field("position", pa.string()),
                                                pa.field("codon", pa.string()),
                                            ]
                                        )
                                    ),
                                ),
                            ]
                        ),
                    ),
                    pa.field(
                        "fr2",
                        pa.struct(
                            [
                                pa.field("num", pa.int64()),
                                pa.field(
                                    "muts",
                                    pa.list_(
                                        pa.struct(
                                            [
                                                pa.field("was", pa.string()),
                                                pa.field("is", pa.string()),
                                                pa.field("raw_position", pa.string()),
                                                pa.field("position", pa.string()),
                                                pa.field("codon", pa.string()),
                                            ]
                                        )
                                    ),
                                ),
                            ]
                        ),
                    ),
                    pa.field(
                        "cdr2",
                        pa.struct(
                            [
                                pa.field("num", pa.int64()),
                                pa.field(
                                    "muts",
                                    pa.list_(
                                        pa.struct(
                                            [
                                                pa.field("was", pa.string()),
                                                pa.field("is", pa.string()),
                                                pa.field("raw_position", pa.string()),
                                                pa.field("position", pa.string()),
                                                pa.field("codon", pa.string()),
                                            ]
                                        )
                                    ),
                                ),
                            ]
                        ),
                    ),
                    pa.field(
                        "fr3",
                        pa.struct(
                            [
                                pa.field("num", pa.int64()),
                                pa.field(
                                    "muts",
                                    pa.list_(
                                        pa.struct(
                                            [
                                                pa.field("was", pa.string()),
                                                pa.field("is", pa.string()),
                                                pa.field("raw_position", pa.string()),
                                                pa.field("position", pa.string()),
                                                pa.field("codon", pa.string()),
                                            ]
                                        )
                                    ),
                                ),
                            ]
                        ),
                    ),
                    pa.field(
                        "cdr3",
                        pa.struct(
                            [
                                pa.field("num", pa.int64()),
                                pa.field(
                                    "muts",
                                    pa.list_(
                                        pa.struct(
                                            [
                                                pa.field("was", pa.string()),
                                                pa.field("is", pa.string()),
                                                pa.field("raw_position", pa.string()),
                                                pa.field("position", pa.string()),
                                                pa.field("codon", pa.string()),
                                            ]
                                        )
                                    ),
                                ),
                            ]
                        ),
                    ),
                    pa.field(
                        "fr4",
                        pa.struct(
                            [
                                pa.field("num", pa.int64()),
                                pa.field(
                                    "muts",
                                    pa.list_(
                                        pa.struct(
                                            [
                                                pa.field("was", pa.string()),
                                                pa.field("is", pa.string()),
                                                pa.field("raw_position", pa.string()),
                                                pa.field("position", pa.string()),
                                                pa.field("codon", pa.string()),
                                            ]
                                        )
                                    ),
                                ),
                            ]
                        ),
                    ),
                ]
            ),
        ),
        pa.field(
            "region_muts_aa",
            pa.struct(
                [
                    pa.field(
                        "fr1",
                        pa.struct(
                            [
                                pa.field("num", pa.int64()),
                                pa.field(
                                    "muts",
                                    pa.list_(
                                        pa.struct(
                                            [
                                                pa.field("was", pa.string()),
                                                pa.field("is", pa.string()),
                                                pa.field("raw_position", pa.string()),
                                                pa.field("position", pa.string()),
                                                pa.field("codon", pa.string()),
                                            ]
                                        )
                                    ),
                                ),
                            ]
                        ),
                    ),
                    pa.field(
                        "cdr1",
                        pa.struct(
                            [
                                pa.field("num", pa.int64()),
                                pa.field(
                                    "muts",
                                    pa.list_(
                                        pa.struct(
                                            [
                                                pa.field("was", pa.string()),
                                                pa.field("is", pa.string()),
                                                pa.field("raw_position", pa.string()),
                                                pa.field("position", pa.string()),
                                                pa.field("codon", pa.string()),
                                            ]
                                        )
                                    ),
                                ),
                            ]
                        ),
                    ),
                    pa.field(
                        "fr2",
                        pa.struct(
                            [
                                pa.field("num", pa.int64()),
                                pa.field(
                                    "muts",
                                    pa.list_(
                                        pa.struct(
                                            [
                                                pa.field("was", pa.string()),
                                                pa.field("is", pa.string()),
                                                pa.field("raw_position", pa.string()),
                                                pa.field("position", pa.string()),
                                                pa.field("codon", pa.string()),
                                            ]
                                        )
                                    ),
                                ),
                            ]
                        ),
                    ),
                    pa.field(
                        "cdr2",
                        pa.struct(
                            [
                                pa.field("num", pa.int64()),
                                pa.field(
                                    "muts",
                                    pa.list_(
                                        pa.struct(
                                            [
                                                pa.field("was", pa.string()),
                                                pa.field("is", pa.string()),
                                                pa.field("raw_position", pa.string()),
                                                pa.field("position", pa.string()),
                                                pa.field("codon", pa.string()),
                                            ]
                                        )
                                    ),
                                ),
                            ]
                        ),
                    ),
                    pa.field(
                        "fr3",
                        pa.struct(
                            [
                                pa.field("num", pa.int64()),
                                pa.field(
                                    "muts",
                                    pa.list_(
                                        pa.struct(
                                            [
                                                pa.field("was", pa.string()),
                                                pa.field("is", pa.string()),
                                                pa.field("raw_position", pa.string()),
                                                pa.field("position", pa.string()),
                                                pa.field("codon", pa.string()),
                                            ]
                                        )
                                    ),
                                ),
                            ]
                        ),
                    ),
                    pa.field(
                        "cdr3",
                        pa.struct(
                            [
                                pa.field("num", pa.int64()),
                                pa.field(
                                    "muts",
                                    pa.list_(
                                        pa.struct(
                                            [
                                                pa.field("was", pa.string()),
                                                pa.field("is", pa.string()),
                                                pa.field("raw_position", pa.string()),
                                                pa.field("position", pa.string()),
                                                pa.field("codon", pa.string()),
                                            ]
                                        )
                                    ),
                                ),
                            ]
                        ),
                    ),
                    pa.field(
                        "fr4",
                        pa.struct(
                            [
                                pa.field("num", pa.int64()),
                                pa.field(
                                    "muts",
                                    pa.list_(
                                        pa.struct(
                                            [
                                                pa.field("was", pa.string()),
                                                pa.field("is", pa.string()),
                                                pa.field("raw_position", pa.string()),
                                                pa.field("position", pa.string()),
                                                pa.field("codon", pa.string()),
                                            ]
                                        )
                                    ),
                                ),
                            ]
                        ),
                    ),
                ]
            ),
        ),
        pa.field("prod", pa.string()),
        pa.field("productivity_issues", pa.string()),
        pa.field("junction_in_frame", pa.string()),
        pa.field("raw_input", pa.string()),
        pa.field("oriented_input", pa.string()),
        pa.field("strand", pa.string()),
        pa.field(
            "germ_alignments_nt",
            pa.struct(
                [
                    pa.field(
                        "var",
                        pa.struct(
                            [
                                pa.field("query", pa.string()),
                                pa.field("germ", pa.string()),
                                pa.field("midline", pa.string()),
                            ]
                        ),
                    ),
                    pa.field(
                        "join",
                        pa.struct(
                            [
                                pa.field("query", pa.string()),
                                pa.field("germ", pa.string()),
                                pa.field("midline", pa.string()),
                            ]
                        ),
                    ),
                    pa.field(
                        "div",
                        pa.struct(
                            [
                                pa.field("query", pa.string()),
                                pa.field("germ", pa.string()),
                                pa.field("midline", pa.string()),
                            ]
                        ),
                    ),
                ]
            ),
        ),
        pa.field(
            "exo_trimming",
            pa.struct(
                [
                    pa.field("var_3", pa.int64()),
                    pa.field("join_5", pa.int64()),
                    pa.field("div_5", pa.int64()),
                    pa.field("div_3", pa.int64()),
                ]
            ),
        ),
        pa.field(
            "junc_nt_breakdown",
            pa.struct(
                [
                    pa.field("v_nt", pa.string()),
                    pa.field("n_nt", pa.string()),
                    pa.field("n1_nt", pa.string()),
                    pa.field("d_nt", pa.string()),
                    pa.field("n2_nt", pa.string()),
                    pa.field("j_nt", pa.string()),
                    pa.field("d_dist_from_cdr3_start", pa.int64()),
                    pa.field("d_dist_from_cdr3_end", pa.int64()),
                ]
            ),
        ),
        pa.field("germline_database", pa.string()),
        pa.field("species", pa.string()),
        pa.field(
            "align_info",
            pa.struct(
                [
                    pa.field("v_start", pa.int64()),
                    pa.field("v_end", pa.int64()),
                    pa.field("d_start", pa.int64()),
                    pa.field("d_end", pa.int64()),
                    pa.field("j_start", pa.int64()),
                    pa.field("j_end", pa.int64()),
                ]
            ),
        ),
        pa.field(
            "j_del",
            pa.list_(
                pa.struct(
                    [
                        pa.field("in_frame", pa.string()),
                        pa.field("length", pa.int64()),
                        pa.field("sequence", pa.string()),
                        pa.field("position", pa.string()),
                        pa.field("codon", pa.string()),
                    ]
                )
            ),
        ),
        pa.field(
            "v_del",
            pa.list_(
                pa.struct(
                    [
                        pa.field("in_frame", pa.string()),
                        pa.field("length", pa.int64()),
                        pa.field("sequence", pa.string()),
                        pa.field("position", pa.string()),
                        pa.field("codon", pa.string()),
                    ]
                )
            ),
        ),
        pa.field(
            "j_ins",
            pa.list_(
                pa.struct(
                    [
                        pa.field("in_frame", pa.string()),
                        pa.field("length", pa.int64()),
                        pa.field("sequence", pa.string()),
                        pa.field("position", pa.string()),
                        pa.field("codon", pa.string()),
                    ]
                )
            ),
        ),
        pa.field(
            "v_ins",
            pa.list_(
                pa.struct(
                    [
                        pa.field("in_frame", pa.string()),
                        pa.field("length", pa.int64()),
                        pa.field("sequence", pa.string()),
                        pa.field("position", pa.string()),
                        pa.field("codon", pa.string()),
                    ]
                )
            ),
        ),
    ]
)



#!/usr/bin/python
# filename: productivity.py

#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from __future__ import absolute_import, division, print_function, unicode_literals

import traceback


def check_productivity(antibody):
    return Productivity(antibody)


class Productivity(object):
    """docstring for Productivity"""

    def __init__(self, antibody):
        super(Productivity, self).__init__()
        self.antibody = antibody
        self.productivity_issues = []
        self._is_productive = None
        self._is_productive_strict = None

    @property
    def is_productive_strict(self):
        """
        Predicts Ab functionality. Similar to ``is_productive``, except sequences with 
        ambiguous nucleotides will be flagged as non-productive.
        """
        if self._is_productive_strict is None:
            try:
                problems = any(
                    [
                        self.stop_codons(self.antibody),
                        self.ambig_codons(self.antibody),
                        self.vdj_disagreement(self.antibody),
                        self.missing_conserved_junc_residues(self.antibody),
                        self.junction_not_in_frame(self.antibody),
                        self.out_of_frame_indels(self.antibody),
                    ]
                )
                if problems:
                    self._is_productive_strict = False
                else:
                    self._is_productive_strict = True
            except:
                self.antibody.exception("PRODUCTIVITY ERROR", traceback.format_exc())
                raise
        return self._is_productive_strict

    @property
    def is_productive(self):
        """
        Predicts Ab functionality. Abs are considered non-productive if they contain
        any of the following:
            - stop codon(s)
            - locus disagreement between V and J genes (an IGHV with an IGKJ, for example)
            - missing conserved junction anchor residues (initial C and terminal W/F)
            - out-of-frame junction
            - indels that are not codon-length
        """
        if self._is_productive is None:
            try:
                problems = any(
                    [
                        self.stop_codons(self.antibody),
                        self.vdj_disagreement(self.antibody),
                        self.missing_conserved_junc_residues(self.antibody),
                        self.junction_not_in_frame(self.antibody),
                        self.out_of_frame_indels(self.antibody),
                    ]
                )
                if problems:
                    self._is_productive = False
                else:
                    self._is_productive = True
            except:
                self.antibody.exception("PRODUCTIVITY ERROR", traceback.format_exc())
                raise
        return self._is_productive

    def stop_codons(self, antibody):
        if "*" in antibody.vdj_aa:
            self.productivity_issues.append("Contains stop codon(s)")
            return True
        return False

    def ambig_codons(self, antibody):
        if "X" in antibody.vdj_aa.upper():
            self.productivity_issues.append("Contains ambiguous codon(s)")
            return True
        return False

    def vdj_disagreement(self, antibody):
        if antibody.v.full[:3] != antibody.j.full[:3]:
            self.productivity_issues.append(
                "V-gene ({}) and J-gene({}) disagree".format(
                    antibody.v.gene, antibody.j.gene
                )
            )
            return True
        if antibody.d:
            if not antibody.d.gene:
                return False
            if antibody.v.full[:3] != antibody.d.full[:3]:
                self.productivity_issues.append(
                    "V-gene ({}) and D-gene({}) disagree".format(
                        antibody.v.gene, antibody.d.gene
                    )
                )
                return True
        return False

    def missing_conserved_junc_residues(self, antibody):
        if not antibody.junction:
            return True
        start = "C"
        end = "W" if antibody.chain == "heavy" else "F"
        if (
            antibody.junction.junction_aa[0] != start
            or antibody.junction.junction_aa[-1] != end
        ):
            self.productivity_issues.append(
                "Junction ({}) lacks conserved start and/or end residue".format(
                    antibody.junction.junction_aa
                )
            )
            return True
        return False

    def junction_not_in_frame(self, antibody):
        if not antibody.junction.in_frame:
            self.productivity_issues.append(
                "Junction ({}) is out of frame".format(antibody.junction.junction_nt)
            )
            return True
        return False

    def out_of_frame_indels(self, antibody):
        problems = 0
        for indels in [
            antibody.v.insertions,
            antibody.j.insertions,
            antibody.v.deletions,
            antibody.j.deletions,
        ]:
            for i in indels:
                if i.in_frame == "no":
                    self.productivity_issues.append(
                        "Contains a frameshift {}: {}".format(
                            i.type, i.abstar_formatted
                        )
                    )
                    problems += 1
        return False if problems == 0 else True



#!/usr/bin/python
# filename: umi.py

#
# Copyright (c) 2023 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import os
import shutil
import sys
import tempfile
from typing import Union, Iterable, Optional

from Bio import SeqIO
from Bio.SeqRecord import SeqRecord

import abutils


class UMI:
    """ """

    def __init__(
        self,
        sequence: Union[str, abutils.Sequence, SeqRecord],
        pattern: str,
        length: Optional[int] = None,
        ignore_strand: bool = False,
        extra_length_for_alignment: int = 25,
    ):
        """ """
        self.length = length
        self.pattern = pattern.strip().upper()
        self.ignore_strand = ignore_strand
        self.extra_length_for_alignment = extra_length_for_alignment
        self.sequence = self.process_sequence(sequence)
        self._leading_aln = None
        self._trailing_aln = None
        self._umi = None
        self._num_mismatches = None

    @property
    def leading(self) -> Optional[str]:
        leading = self.pattern.split("[UMI]")[0]
        return leading if leading else None

    @property
    def trailing(self) -> Optional[str]:
        trailing = self.pattern.split("[UMI]")[-1]
        return trailing if trailing else None

    @property
    def leading_aln(self) -> abutils.SSWAlignment:
        if self._leading_aln is None:
            if self.leading is not None:
                self._leading_aln = self.align(self.leading, self.sequence)
        return self._leading_aln

    @property
    def trailing_aln(self) -> abutils.SSWAlignment:
        if self._trailing_aln is None:
            if self.trailing is not None:
                self._trailing_aln = self.align(self.trailing, self.sequence)
        return self._trailing_aln

    @property
    def umi(self) -> Optional[str]:
        if self._umi is None:
            self._umi = self.get_umi()
        return self._umi

    @property
    def num_mismatches(self) -> int:
        """
        Returns the number of mismatches between the pattern and sequence.
        """
        if self._num_mismatches is None:
            self._num_mismatches = self.get_mismatches()
        return self._num_mismatches

    def process_sequence(self, sequence) -> str:
        """
        Processes the raw sequence:
          * reverse complement if length is negative
          * truncate sequence prior to pattern alignment

        Parameters
        ----------
        sequence : str, Sequence, or SeqRecord
            Raw sequence

        Returns
        -------
        truncated : str
        """
        s = abutils.Sequence(sequence)
        length_for_aln = (
            len(self.pattern.replace("[UMI]", "")) + self.extra_length_for_alignment
        )
        if self.length is not None:
            length_for_aln += self.length
        if self.length < 0 and not self.ignore_strand:
            return s.reverse_complement[:length_for_aln]
        else:
            return s[:length_for_aln]

    def align(
        self, pattern: str, sequence: Union[abutils.Sequence, str]
    ) -> abutils.SSWAlignment:
        """
        Performs local alignment between a pattern and sequence.
        """
        aln = abutils.tl.local_alignment(
            pattern, sequence, gap_open=-20, gap_extend=-20
        )
        return aln

    def get_mismatches(self) -> Optional[int]:
        """
        Returns the total number of mismatches between the pattern(s)
        and the input sequence.
        """
        total = len(self.pattern.replace("[UMI]", ""))
        matches = 0
        for aln in [self.leading_aln, self.trailing_aln]:
            if aln is not None:
                matches += aln.alignment_midline.count("|")
        return total - matches

    def get_umi(self) -> Optional[str]:
        """
        Parses UMI from the input sequence using the UMI pattern(s).
        """
        ## something like "ATGC[UMI]"
        ## we need to take `length` residues following the end of the leading alignment
        if self.trailing is None and self.leading is not None:
            start = self.leading_aln.target_end + 1
            end = start + self.length
            if end > len(self.sequence):
                return None
        ## something like "[UMI]ATGC"
        ## we need to take `length` residues preceeding the start of the trailing alignment
        elif self.leading is None and self.trailing is not None:
            end = self.trailing_aln.target_begin
            if self.length is not None:
                start = end - self.length
            else:
                start = 0
            ## if self.length is longer than the region preceeding the start of the
            ## trailing alignment, return None because we can't parse a complete UMI
            if start < 0:
                return None
        else:
            start = self.leading_aln.target_end + 1
            end = self.trailing_aln.target_begin
        return self.sequence[start:end]


def parse_umis(
    sequences: Union[str, Iterable],
    output_file: Optional[str] = None,
    pattern: Union[str, Iterable, None] = None,
    length: Union[int, Iterable, None] = None,
    allowed_mismatches: Optional[int] = None,
    extra_length_for_alignment: int = 25,
    ignore_strand: bool = False,
    fmt: str = "fasta",
    sequence_key: str = "raw_input",
    id_key: str = "sequence_id",
):
    """
    Parses unique molecular identifiers (UMIs) from bulk NGS sequencing data.

    Parameters
    ----------
    sequences : str or iterable
        Can be one of several things:
            1. path to a FASTA- or FASTQ-formatted file
            2. a list of BioPython ``SeqRecord`` objects
            3. a list of abutils ``Sequence`` objects
            4. a list of lists/tuples, of the format ``[sequence_id, sequence]``

    output_file : str, default=None
        Path to an output file. Required if `sequences` is not a
        file. If `sequences` is a file and `output` is not provided, UMI parsing is
        done in-place and the `sequences` file is updated.

    pattern : str or iterable, default=None
        Pattern (or iterable of patterns) for identifying the location of the UMI,
        or the name of a built-in pattern. Built-in options include ``"smartseq-human-bcr"``.
        Patterns may optionally contain leading and/or trailing conserved regions, with
        the UMI position within the pattern represented by ``"[UMI]"``. As an example,
        the built-in pattern for SmartSeq-BCR UMIs is::

            "[UMI]TCAGCGGGAAGACATT"

        which would be a UMI sequence followed immediately by ``"TCAGCGGGAAGACATT"``.
        By default, the pattern is matched on the 5' -> 3' strand. This allows
        users to more easily construct patterns from their amplification primers without
        needing to worry about reverse-complementing patterns for UMIs at the 3' end of the
        input sequence. To override this, set `ignore_strand` to ``True``. If `pattern`
        is not provided, UMIs will be parsed using only `length`, starting at the start
        or end of the sequence.

        .. note::

            The UMIs for all `patterns` that meet the `allowed_mismatches` criteria in
            the conserved leading/trailing portions will be concatenated into the final
            UMI. This allows the use of multiple `patterns` for either:
              * different patterns designed to match one of several different types of
                sequences in a heterogeneous sample. For example, if heavy, kappa and
                lambda primers each have different conserved regions flanking the UMI
                and the input file contains a mix of heavy, kappa and lambda chains,
                supplying all patterns (assuming they're sufficiently different from
                each other) will allow parsing of UMIs from all chains.
              * different patterns for sequences that contain multiple UMIs, either
                at opposite ends of the sequence or on the same end of the sequence
                but in different locations and with different conserved flanking regions.

    length : int or iterable, default=None
        Length of the UMI sequence, or iterable of UMI lengths. If multiple lengths are
        provided, there must be an equal number of `patterns`, and they must be in the
        same order (the first `pattern` should correspond to the first UMI `length`). If
        `length` is positive, the UMI will be parsed from the start of the sequence. If
        `length` is negative, the UMI will be parsed from the end of the sequence. If multiple
        `patterns` are provided with a single `length`, that `length` will be used for all
        `patterns`. Required if `pattern` does not have a conserved trailing region.
        If `length` is not provided and a trailing conserved region is present in `pattern`,
        the entire portion of `sequence` preceeding the trailing region will be parsed as the
        UMI. Ignored if both leading and trailing sequences are present in `pattern`,
        as the entire region between the conserved flanking regions will be parsed as
        the UMI regardless of what `length` is provided.





    """
    # lengths and patterns
    if isinstance(pattern, str):
        if pattern.lower() in BUILTIN_PATTERNS:
            name = pattern.lower()
            patterns = _get_patterns(name)
            length = _get_lengths(name)
            if allowed_mismatches is None:
                allowed_mismatches = _get_allowed_mismatches(name)
        else:
            patterns = [pattern]
    else:
        patterns = pattern
    if isinstance(length, int) or length is None:
        lengths = [length] * len(patterns)
    else:
        lengths = length
    if len(patterns) != len(lengths):
        err = "\nERROR: if pattern and length are both iterables,"
        err += " they must be the same length.\n"
        err += f"  - pattern has {len(patterns)} elements: {', '.join(patterns)}\n"
        err += f"  - length has {len(lengths)} elements: {', '.join(lengths)}\n"
        print(err)
        sys.exit()
    # mismatches
    if allowed_mismatches is None:
        allowed_mismatches = 1
    # validate format
    fmt = fmt.lower()
    if fmt not in ["fasta", "fastq"]:
        err = "\nERROR: format must be either 'fasta' or 'fastq'. "
        err += f"You provided {fmt}.\n"
        print(err)
        sys.exit()
    # input processing
    if isinstance(sequences, str) and os.path.isfile(sequences):
        output_file = _parse_umis_from_file(
            input_file=sequences,
            patterns=patterns,
            lengths=lengths,
            output_file=output_file,
            allowed_mismatches=allowed_mismatches,
            extra_length_for_alignment=extra_length_for_alignment,
            ignore_strand=ignore_strand,
            fmt=fmt,
        )
    else:
        if output_file is None:
            err = "\nERROR: if input is not a FASTA- or FASTQ-formatted file, "
            err += f"output_file must be provided.\n"
            print(err)
            sys.exit()
        output_file = _parse_umis_from_sequences(
            sequences=sequences,
            patterns=patterns,
            lengths=lengths,
            output_file=output_file,
            allowed_mismatches=allowed_mismatches,
            extra_length_for_alignment=extra_length_for_alignment,
            ignore_strand=ignore_strand,
            fmt=fmt,
            sequence_key=sequence_key,
            id_key=id_key,
        )


def _parse_umis_from_file(
    input_file: str,
    patterns: Iterable,
    lengths: Iterable,
    output_file: Optional[str] = None,
    allowed_mismatches: int = 1,
    extra_length_for_alignment: int = 25,
    ignore_strand: bool = False,
    fmt: str = "fasta",
) -> str:
    """
    Parses UMIs from a FASTA- or FASTQ-formatted file.
    """
    # set up output file
    inplace = False
    if output_file is None:
        output_file = tempfile.NamedTemporaryFile(delete=False).name
        inplace = True
    with open(output_file, "w") as ofile:
        with open(input_file, "r") as ifile:
            for s in SeqIO.parse(ifile, fmt.lower()):
                s = abutils.Sequence(s)
                umi_list = []
                for p, l in zip(patterns, lengths):
                    u = UMI(
                        sequence=s,
                        pattern=p,
                        length=l,
                        extra_length_for_alignment=extra_length_for_alignment,
                        ignore_strand=ignore_strand,
                    )
                    if u.num_mismatches <= allowed_mismatches:
                        umi_list.append(u.umi)
                # if we couldn't find a UMI, skip the sequence
                umi_list = [u for u in umi_list if u is not None]
                if not umi_list:
                    continue
                umi = "+".join(umi_list)
                s.id = f"{s.id}_{umi}"
                if fmt.lower() == "fastq":
                    fstring = s.fastq + "\n"
                else:
                    fstring = s.fasta + "\n"
                ofile.write(fstring)
    if inplace:
        output_file = shutil.move(output_file, input_file)
    return output_file


def _parse_umis_from_sequences(
    sequences: Iterable,
    patterns: Iterable,
    lengths: Iterable,
    output_file: str,
    allowed_mismatches: int = 1,
    extra_length_for_alignment: int = 25,
    ignore_strand: bool = False,
    fmt: str = "fasta",
    sequence_key: str = "raw_input",
    id_key: str = "sequence_id",
) -> str:
    """
    Parses UMIs from a list of sequences.
    """
    with open(output_file, "w") as ofile:
        for s in sequences:
            if isinstance(s, abutils.Sequence):
                if sequence_key in s:
                    s.sequence = s[sequence_key]
                if id_key in s:
                    s.id = s[id_key]
            s = abutils.Sequence(s)
            umi_list = []
            for p, l in zip(patterns, lengths):
                u = UMI(
                    sequence=s,
                    pattern=p,
                    length=l,
                    extra_length_for_alignment=extra_length_for_alignment,
                    ignore_strand=ignore_strand,
                )
                if u.num_mismatches <= allowed_mismatches:
                    umi_list.append(u.umi)
            # if we couldn't find a UMI, skip the sequence
            umi_list = [u for u in umi_list if u is not None]
            if not umi_list:
                continue
            umi = "+".join(umi_list)
            s.id = f"{s.id}_{umi}"
            if fmt.lower() == "fastq":
                fstring = s.fastq + "\n"
            else:
                fstring = s.fasta + "\n"
            ofile.write(fstring)
    return output_file


def _get_patterns(name):
    p = BUILTIN_PATTERNS.get(name, None)
    if p is not None:
        return p.get("pattern", None)


def _get_lengths(name):
    p = BUILTIN_PATTERNS.get(name, None)
    if p is not None:
        return p.get("length", None)


def _get_allowed_mismatches(name):
    p = BUILTIN_PATTERNS.get(name, None)
    if p is not None:
        return p.get("allowed_mismatches", None)


# ---------------------------
#        PATTERNS
# ---------------------------

# Takara's Smart-Seq Human BCR kit
# https://www.takarabio.com/products/next-generation-sequencing/immune-profiling/human-repertoire/smart-seq-human-bcr-with-umis
SMARTSEQ_HUMAN_BCR_PATTERN = [
    "[UMI]TCAGCGGGAAGACATT",
    "[UMI]GGGCGGATGGACTACC",
    "[UMI]CCGATGGGCGCTTGGT",
    "[UMI]GGAACACATGCGGAGC",
]
SMARTSEQ_HUMAN_BCR_LENGTH = 12
SMARTSEQ_HUMAN_BCR_MISMATCH = 2


BUILTIN_PATTERNS = {
    "smartseq-human-bcr": {
        "pattern": SMARTSEQ_HUMAN_BCR_PATTERN,
        "length": SMARTSEQ_HUMAN_BCR_LENGTH,
        "allowed_mismatches": SMARTSEQ_HUMAN_BCR_MISMATCH,
    },
}



#!/usr/bin/python
# filename: isotype.py

#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import os
import traceback

from Bio import SeqIO
from Bio.Seq import Seq

from abutils.core.sequence import Sequence
from abutils.utils import log
from abutils.utils.alignment import local_alignment
from abutils.utils.decorators import lazy_property

from ..core.germline import get_germline_database_directory


def get_isotype(antibody):
    try:
        germ_dir = get_germline_database_directory(antibody.germ_db)
        isotype_file = os.path.join(germ_dir, "isotypes/isotypes.fasta")
        isotype_seqs = [
            Sequence(s) for s in SeqIO.parse(open(isotype_file, "r"), "fasta")
        ]
        return Isotype(antibody, isotype_seqs)
    except:
        antibody.exception("ISOTYPING ERROR", traceback.format_exc())


# def get_isotype(vdj):
#     logger = log.get_logger(__name__)
#     try:
#         mod_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
#         isotype_file = os.path.join(mod_dir, 'ssw/isotypes/{}_isotypes.fasta'.format(vdj.species))
#         isotype_seqs = [Sequence(s) for s in SeqIO.parse(open(isotype_file, 'r'), 'fasta')]
#         isotype_seqs += [Sequence((s.id, s.reverse_complement)) for s in isotype_seqs]
#         return Isotype(vdj, isotype_seqs)
#     except:
#         logger.debug('ISOTYPE ERROR: {}\t{}'.format(vdj.id, vdj.raw_query))
#         logger.debug(traceback.format_exc())


class Isotype(object):
    """docstring for Isotype"""

    def __init__(self, antibody, isotype_seqs):
        super(Isotype, self).__init__()
        self._alignments = self._get_alignments(antibody, isotype_seqs)
        self.alignment = self._alignments[0]

    @lazy_property
    def isotype(self):
        if self.normalized_score < 2:
            return "unknown"
        return self.alignment.target.id

    @lazy_property
    def score(self):
        return self.alignment.score

    @lazy_property
    def normalized_score(self):
        return float(self.alignment.score) / len(self.alignment)

    def _get_alignments(self, antibody, isotype_seqs):
        query_region = self._get_isotype_query_region(antibody)
        if len(query_region) == 0:
            return []
        alignments = local_alignment(
            query_region,
            targets=isotype_seqs,
            gap_open_penalty=22,
            gap_extend_penalty=1,
        )
        return sorted(alignments, key=lambda x: x.score, reverse=True)

    def _get_isotype_query_region(self, antibody):
        aln = local_alignment(antibody.vdj_nt, antibody.oriented_input)
        return antibody.oriented_input[aln.target_end :]



#!/usr/bin/env python
# filename: blastn.py

#
# Copyright (c) 2016 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

import os
import platform
from tempfile import NamedTemporaryFile
import traceback

from Bio import SeqIO
from Bio.Blast.Applications import NcbiblastnCommandline
from Bio.Blast import NCBIXML

from abutils.core.sequence import Sequence
from abutils.utils.alignment import local_alignment

from .assigner import BaseAssigner
from ..core.germline import GermlineSegment
from ..core.vdj import VDJ


class Blastn(BaseAssigner):
    """
    docstring for Blastn
    """

    def __init__(self, db_name, receptor):
        super(Blastn, self).__init__(db_name, receptor)

    def __call__(self, sequence_file, file_format):
        with open(sequence_file, "r") as sequence_handle:
            seqs = [Sequence(s) for s in SeqIO.parse(sequence_handle, file_format)]
        vdjs = []

        # if the input file is FASTQ-formatted, need to convert it to FASTA for BLASTn to work
        if file_format == "fastq":
            with open(sequence_file, "w") as handle:
                handle.write("\n".join(s.fasta for s in seqs))

        # assign V-genes
        vblast_records = self.blast(sequence_file, "V")
        # if there aren't any vblast_records, that means that none of the
        # sequences in the input file contained sequences with a significant
        # match to any germline V-gene. These are likely all non-antibody sequences.
        if not vblast_records:
            vdjs = [VDJ(seq) for seq in seqs]
            for vdj in vdjs:
                vdj.log(
                    "V-GENE ASSIGNMENT ERROR:",
                    "No variable gene was found.",
                    "Query sequence does not appear to contain a rearranged adaptive immune receptor.",
                )
            self.unassigned = vdjs
            return
        jquery_seqs = []
        for seq, vbr in zip(seqs, vblast_records):
            try:
                germ = self.process_blast_record(vbr)
                vdj = VDJ(seq, v=germ)
                self.orient_query(vdj, vbr)
                jquery = self.get_jquery_sequence(vdj.oriented, vbr)
                # only try to find J-genes if there's a minimum of 10 nucleotides
                # remaining after removal of the V-gene alignment
                if len(jquery) >= 10:
                    vdjs.append(vdj)
                    jquery_seqs.append(jquery)
                # abort VDJ assignment if the J-gene query sequence is too short
                else:
                    vdj = VDJ(seq)
                    vdj.log(
                        "J-GENE QUERY ERROR:",
                        "Query sequence for J-gene assignment is too short.",
                    )
                    vdj.log("")
                    query = vbr.alignments[0].hsps[0].query
                    subject = vbr.alignments[0].hsps[0].sbjct
                    vdj.log(" QUERY :", query)
                    vdj.log(
                        "        ",
                        "".join(
                            ["|" if q == s else " " for q, s in zip(query, subject)]
                        ),
                    )
                    vdj.log("SUBJECT:", subject)
                    vdj.log("")
                    vdj.log("J-QUERY SEQUENCE:", jquery.sequence)
                    self.unassigned.append(vdj)
            except:
                vdj = VDJ(seq)
                vdj.exception("V-GENE ASSIGNMENT ERROR", traceback.format_exc())
                self.unassigned.append(vdj)

        # assign J-genes
        _vdjs = []
        dquery_seqs = []
        jblast_infile = self.build_jblast_input(jquery_seqs)
        jblast_records = self.blast(jblast_infile, "J")
        for vdj, jquery, jbr in zip(vdjs, jquery_seqs, jblast_records):
            try:
                # germ = self.process_blast_record(jbr)
                germ = self.process_blast_record(jbr, match_locus=vdj.v.full[:3])
                vdj.j = germ
                # sanity check to make sure there's not an obvious problem with the V/J
                # assignments (likely due to poor germline matches to a non-antibody sequence)
                if vdj.v.chain != vdj.j.chain:
                    vdj.log(
                        "GERMLINE ASSIGNMENT ERROR:",
                        "V-gene ({}) and J-gene ({}) chains do not match".format(
                            vdj.v.chain, vdj.j.chain
                        ),
                    )
                    vdj.log(
                        "TOP J-GENE ASSIGNMENTS:",
                        "\n  - "
                        + "\n  - ".join(
                            [f"{j.full} ({j.assigner_score})" for j in vdj.j.others]
                        ),
                    )
                    self.unassigned.append(vdj)
                    continue
                dquery = self.get_dquery_sequence(jquery, jbr)
                dquery_seqs.append(dquery)
                _vdjs.append(vdj)
            except:
                vdj.exception("V-GENE ASSIGNMENT ERROR", traceback.format_exc())
                self.unassigned.append(vdj)
        os.unlink(jblast_infile)
        vdjs = _vdjs

        # assign D-genes
        _vdjs = []
        for vdj, dquery in zip(vdjs, dquery_seqs):
            if all([vdj.v.chain in ["heavy", "beta", "delta"], dquery]):
                try:
                    germ = self.assign_dgene(dquery)
                    vdj.d = germ
                except:
                    vdj.exception("D-GENE ASSIGNMENT ERROR:", traceback.format_exc())
                    self.unassigned.append(vdj)
                    continue
            _vdjs.append(vdj)
        self.assigned = _vdjs

    # @property
    # def name(self):
    #     return 'blastn'

    def blast(self, seq_file, segment):
        """
        Runs BLASTn against an antibody germline database.

        Args:
        -----

            seq_file (str): Path to a FASTA-formatted file of input sequences.

            segment (str): Germline segment to query. Options are ``V`` and ``J``.
        """
        blast_path = os.path.join(
            self.binary_directory, "blastn_{}".format(platform.system().lower())
        )
        blast_db_path = os.path.join(
            self.germline_directory, "blast/{}".format(segment.lower())
        )
        blastout = NamedTemporaryFile(delete=False, mode="r")
        blastn_cmd = NcbiblastnCommandline(
            cmd=blast_path,
            db=blast_db_path,
            query=seq_file,
            out=blastout.name,
            outfmt=5,
            dust="no",
            word_size=self._word_size(segment),
            max_target_seqs=10,
            evalue=self._evalue(segment),
            reward=self._match_reward(segment),
            penalty=self._mismatch_penalty(segment),
            gapopen=self._gap_open(segment),
            gapextend=self._gap_extend(segment),
        )
        stdout, stderr = blastn_cmd()
        blast_records = [br for br in NCBIXML.parse(blastout)]
        os.unlink(blastout.name)
        return blast_records

    def assign_dgene(self, seq):
        db_file = os.path.join(self.germline_directory, "ungapped/d.fasta")
        with open(db_file, "r") as db_handle:
            germs = [Sequence(s) for s in SeqIO.parse(db_handle, "fasta")]
            rc_germs = [Sequence(s.reverse_complement, id=s.id) for s in germs]
            germs.extend(rc_germs)
        alignments = local_alignment(seq, targets=germs, gap_open=-20, gap_extend=-2)
        alignments.sort(key=lambda x: x.score, reverse=True)
        all_gls = [a.target.id for a in alignments]
        if "__" in all_gls[0]:
            species = all_gls[0].split("__")[-1].replace("-", " ")
            # all_gls = [gl.split('__')[0] for gl in all_gls]
        else:
            species = self.db_name
        all_scores = [a.score for a in alignments]
        if not all([all_gls, all_scores]):
            return None
        top_gl = all_gls[0]
        top_score = all_scores[0]
        others = [
            GermlineSegment(germ, species, self.db_name, self.receptor, score=score)
            for germ, score in zip(all_gls[1:6], all_scores[1:6])
        ]
        return GermlineSegment(
            top_gl,
            species,
            self.db_name,
            self.receptor,
            score=top_score,
            others=others,
            assigner_name=self.name,
        )

    def process_blast_record(
        self, blast_record: str, match_locus: str = None
    ) -> GermlineSegment:
        """
        Processes a single BLAST record.

        Parameters
        ----------
        blast_record : str
            A single BLAST record, output by BLASTn.

        match_locus : str, default=None
            Locus of a paired germline gene to match. Used when processing 
            BLAST records for J gene assignemnts, to ensure that the locus 
            of the J gene matches the already assigned V gene. Should be a 
            three-character string. Options are ``"IGH"``, ``"IGK"``, ``"IGL"``,
            ``"TRA"``, ``"TRB"``, ``"TRD"`` or ``"TRG"``.

        Returns
        -------
        germline : GermlineSegment
        
        """
        all_gls = [a.title.split()[0] for a in blast_record.alignments]
        if match_locus is not None:
            matching_gls = [gl for gl in all_gls if gl[:3] == match_locus]
            if matching_gls:
                all_gls = matching_gls
            #     n_others = min(5, len(all_gls))
            # else:
            #     n_others = len(all_gls)
        else:
            n_others = min(5, len(all_gls))
        if "__" in all_gls[0]:
            species = all_gls[0].split("__")[-1].replace("-", " ")
            # all_gls = [gl.split('__')[0] for gl in all_gls]
        else:
            species = self.db_name
        all_scores = [a.hsps[0].bits for a in blast_record.alignments]
        top_gl = all_gls[0]
        top_score = all_scores[0]
        others = [
            GermlineSegment(germ, species, self.db_name, self.receptor, score=score)
            for germ, score in zip(all_gls[1:], all_scores[1:])
        ]
        n_others = min(5, len(others))
        return GermlineSegment(
            top_gl,
            species,
            self.db_name,
            self.receptor,
            score=top_score,
            others=others[:n_others],
            assigner_name=self.name,
        )

    @staticmethod
    def orient_query(vdj, vbr):
        hsp = vbr.alignments[0].hsps[0]
        # BLASTn always reverse complements the Subject sequence, never the query.
        # To determine whether the input sequence is the reverse complement, check
        # to see if the Subject sequence was reverse-complemented by BLASTn
        if hsp.sbjct_start > hsp.sbjct_end:
            vdj.oriented = Sequence(vdj.sequence.reverse_complement, id=vdj.sequence.id)
            vdj.v.strand = "-"
        else:
            vdj.v.strang = "+"

    @staticmethod
    def get_jquery_sequence(seq, vbr):
        hsp = vbr.alignments[0].hsps[0]
        # check to see if the raw input was reverse-complemented
        if hsp.sbjct_start > hsp.sbjct_end:
            # since the BLASTn alignment was done on the raw input
            # (which has since been reverse complemented), we need
            # to take the portion of the sequence that was 5' of the alignment
            # with the raw input (which is the 3' end of the correctly oriented sequence)
            return Sequence(seq[-hsp.query_start :], id=seq.id)
        else:
            return Sequence(seq[hsp.query_end :], id=seq.id)

    @staticmethod
    def get_dquery_sequence(seq, jbr):
        query_start = jbr.alignments[0].hsps[0].query_start - 1
        return Sequence(seq[:query_start], id=seq.id)

    @staticmethod
    def build_jblast_input(jseqs):
        jblast_input = NamedTemporaryFile(delete=False, mode="w")
        jblast_input.write("\n".join([s.fasta for s in jseqs]))
        jblast_input.close()
        return jblast_input.name

    @staticmethod
    def _word_size(segment):
        word_sizes = {"V": 11, "D": 4, "J": 7}
        return word_sizes[segment]

    @staticmethod
    def _gap_open(segment):
        gap_open = {"V": 5, "D": 4, "J": 5}
        return gap_open[segment]

    @staticmethod
    def _gap_extend(segment):
        gap_extend = {"V": 2, "D": 2, "J": 2}
        return gap_extend[segment]

    @staticmethod
    def _match_reward(segment):
        match = {"V": 1, "D": 1, "J": 1}
        return match[segment]

    @staticmethod
    def _mismatch_penalty(segment):
        mismatch = {"V": -1, "D": -1, "J": -1}
        return mismatch[segment]

    @staticmethod
    def _evalue(segment):
        evalue = {"V": 1, "D": 100000, "J": 1000}
        return evalue[segment]



#!/usr/bin/env python
# filename: germline.py

#
# Copyright (c) 2016 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from __future__ import absolute_import, division, print_function, unicode_literals

import math
import os
import traceback
from typing import Optional, Union, Iterable

import parasail

from Bio import SeqIO
from Bio.Seq import Seq

from abutils.core.sequence import Sequence
from abutils.utils.alignment import global_alignment, local_alignment
from abutils.utils.codons import codon_lookup
from abutils.utils.decorators import lazy_property

# from .antibody import Antibody
from ..utils import indels
from ..utils.mixins import LoggingMixin


class GermlineSegment(LoggingMixin):
    """
    docstring for Germline

    Parameters
    ----------
    raw : str
        Raw germline assignment, which includes the species name.

    species : str
        Species from which the germline gene was derived. Choices include
        'human', 'macaque', mouse', rabbit'.

    db_name : str
        Name of the germline database, as a string.

    receptor : str
        Receptor name. Options are ``"bcr"`` and ``"tcr"``.

    score : int or float, default=None
        Score of the top germline alignment. Optional. If ``score`` is not
        provided, the realignment score will be used instead.

    strand : str, default=None
        Strand of the alignment. Options are ``'+'`` for the positive strand
        (meaning the germline gene and the query sequence are in the same orientation)
        or ``'-'`` for the negative strand (query sequence is the reverse complement of
        the germline gene).

    others : list(GermlineSegment), default=None
        A list of additional high scoring germline genes. Each element of the list
        should be a ``GermlineSegment``.

    assigner_name : str, default=None
        The assigner name. Will be converted to lowercase. If not provided,
        `assigner_name` will be set to ``'unknown'``.

    initialize_log : bool, default=True
        Whether or not to initialize logging.

    """

    def __init__(
        self,
        raw: str,
        species: str,
        db_name: str,
        receptor: str,
        score: Optional[Union[float, int]] = None,
        strand: Optional[str] = None,
        others: Optional[Iterable] = None,
        assigner_name: Optional[str] = None,
        initialize_log: bool = True,
    ):
        super(GermlineSegment, self).__init__()
        LoggingMixin.__init__(self)
        self.raw_assignment = raw
        self.full = raw.split("__")[0]
        self.species = species.lower()
        self.db_name = db_name
        self.receptor = receptor.lower()
        self.assigner_score = score
        self.strand = strand
        self.others = others
        self.gene_type = self.full[3].upper()
        self.assigner = (
            assigner_name.lower() if assigner_name is not None else "unknown"
        )
        self._family = None
        self._gene = None
        self._chain = None

        # Optional properties for assigners to populate.
        # If populated by an assigner, they will be used to
        # force realignment to these parameters.
        #
        # Note that the query_start/query_end positions should
        # be 0-indexed and apply to the full query sequence. So if the sequence
        # is truncated following V-gene assignment and the truncated
        # sequence is submitted for J-gene assignment, make sure to
        # adjust the positions for the J-gene so that the numberings apply to
        # the complete input sequence, not the truncated sequence that was
        # used to identify the J-gene.
        #
        # Also note that because the end position of Python's slicing operator
        # is exclusive, to get the aligned sequence from the oriented_input, you
        # need to add one to the query_end (so that the slice includes query_end)
        self.query_start = None
        self.query_end = None
        self.germline_start = None
        self.germline_end = None

        # initialize log
        if initialize_log:
            self.initialize_log()

        # These properties are populated by AbStar.
        # Assigners don't need to populate these (and they'll be overwritten
        # by AbStar if an assigner does populate them).
        self.score = None
        self.realignment = None
        self.raw_query = None
        self.raw_germline = None
        self.query_alignment = None
        self.germline_alignment = None
        self.alignment_midline = None
        self.alignment_length = None
        self.alignment_reading_frame = None  # 0-based, so if alignment start is the start of a codon, reading frame will be 0
        self.imgt_germline = None
        self.imgt_gapped_alignment = None
        self.imgt_nt_positions = []
        self.imgt_aa_positions = []
        self._imgt_position_from_raw = {}
        self._raw_position_from_imgt = {}
        self._initial_correct_imgt_nt_position_from_imgt = None
        self._initial_correct_imgt_aa_position_from_imgt = None
        self._correct_imgt_nt_position_from_imgt = None
        self._correct_imgt_aa_position_from_imgt = None
        self.fs_indel_adjustment = 0
        self.nfs_indel_adjustment = 0
        self.has_insertion = "no"
        self.has_deletion = "no"
        self._insertions = None
        self._deletions = None
        self.coding_region = None
        self.aa_sequence = None
        self.regions = None

    @property
    def family(self):
        if self._family is None:
            if "-" in self.full:
                self._family = self.full.split("-")[0]
        return self._family

    @family.setter
    def family(self, family):
        self._family = family

    @property
    def gene(self):
        if self._gene is None:
            if "*" in self.full:
                self._gene = self.full.split("*")[0]
        return self._gene

    @gene.setter
    def gene(self, gene):
        self._gene = gene

    @property
    def chain(self):
        if self._chain is None:
            c = {
                "H": "heavy",
                "K": "kappa",
                "L": "lambda",
                "A": "alpha",
                "B": "beta",
                "G": "gamma",
                "D": "delta",
            }
            self._chain = c.get(self.full[2], None)
        return self._chain

    @property
    def insertions(self):
        if self._insertions is None:
            return []
        return self._insertions

    @insertions.setter
    def insertions(self, insertions):
        self._insertions = insertions

    @property
    def deletions(self):
        if self._deletions is None:
            return []
        return self._deletions

    @deletions.setter
    def deletions(self, deletions):
        self._deletions = deletions

    def correct_imgt_nt_position_from_imgt(self, position):
        if self._correct_imgt_nt_position_from_imgt is not None:
            p = self._correct_imgt_nt_position_from_imgt.get(position, None)
            if p is not None:
                return p
        if self._initial_correct_imgt_nt_position_from_imgt is not None:
            p = self._initial_correct_imgt_nt_position_from_imgt.get(position, None)
            if p is not None:
                return p
        return None

    def correct_imgt_aa_position_from_imgt(self, position):
        if self._correct_imgt_aa_position_from_imgt is not None:
            p = self._correct_imgt_aa_position_from_imgt.get(position, None)
            if p is not None:
                return p
        if self._initial_correct_imgt_aa_position_from_imgt is not None:
            p = self._initial_correct_imgt_aa_position_from_imgt.get(position, None)
            if p is not None:
                return p
        return None

    def initialize_log(self):
        log = []
        log.append("GERMLINE: {}".format(self.full))
        self._log = log

    def realign_germline(
        self,
        antibody,
        query_start: Optional[int] = None,
        query_end: Optional[int] = None,
    ) -> None:
        """
        Due to restrictions on the available scoring parameters in BLASTn, incorrect truncation
        of the v-gene alignment can occur. This function re-aligns the query sequence with
        the identified germline variable gene using more appropriate alignment parameters.

        Parameters
        ----------
        antibody : Antibody
            The ``Antibody`` object to which this ``GermlineSegment`` object is attached.

        query_start: int, default=None
            Position in the input sequence at which the re-alignment should start. If not provided,
            the start of the input sequence is used.

        query_end: int, default=None
            Position in the input sequence at which the re-alignment should end. If not provided,
            the end of the input sequence is used.

        """
        oriented_input = antibody.oriented_input
        germline_seq = self._get_germline_sequence_for_realignment(antibody)
        if germline_seq is None:
            antibody.log("GET GERMLINE SEQUENCE ERROR")
            antibody.log("RAW ASSIGNMENT:", self.raw_assignment)
            antibody.log("GERMLINE GENE:", self.full)
        aln_params = self._realignment_scoring_params(self.gene_type)
        # if the alignment start/end positions have been annotated by the assigner,
        # force re-alignment using those parameters
        if all(
            [
                x is not None
                for x in [
                    self.query_start,
                    self.query_end,
                    self.germline_start,
                    self.germline_end,
                ]
            ]
        ):
            query = oriented_input.sequence[self.query_start : self.query_end]
            germline = germline_seq[self.germline_start : self.germline_end]
            alignment = global_alignment(query, germline, **aln_params)
        # use local alignment to determine alignment start/end positions if
        # they haven't already been determined by the assigner
        else:
            query = oriented_input.sequence[query_start:query_end]
            if len(query) == 0:
                antibody.log("GERMLINE REALIGNMENT ERROR: query sequence is empty")
                antibody.log("REALIGNMENT QUERY SEQUENCE:", query)
                antibody.log("QUERY START:", query_start)
                antibody.log("QUERY END:", query_end)
                return
            alignment = local_alignment(query, germline_seq, **aln_params)
            # fix for a fairly rare edge case where coincidental matching to 2-3 residues at the extreme
            # 3' end of K/L germline V genes can result in incorrect identification of the
            # end of the V gene region (making the called V region far too long and, in some cases,
            # extending beyond the junction and into FR4). What we do here is drop the last 2 nucleotides
            # of the aligned germline and re-align to see whether that substantialy truncates the resulting
            # alignment (by at least 2 additional nucleotides). If so, we use the new alignment instead.
            if self.gene_type == "V" and self.chain in ["kappa", "lambda"]:
                germline_trunc = germline_seq[: alignment.target_end - 2]
                alignment2 = local_alignment(query, germline_trunc, **aln_params)
                if alignment.query_end - alignment2.query_end >= 4:
                    alignment2.raw_target = (
                        alignment.raw_target
                    )  # swap the truncated germline with the full one
                    alignment = alignment2
            # occasionally, a small number of mismatches can cause a sequence which encodes a
            # complete germline gene region to be incorrectly truncated at the 5' end (V genes) or 3' end (J genes)
            # this can cause a problem when synthesizing mAbs for recombinant expression, as the annotated VDJ
            # sequence is incomplete
            #
            # here we check to see if the alignment was truncated and whether there are additional bases in the
            # input sequence that should be included in the alignment
            # if so, we force the alignment to extend to the entire germline gene region
            if self.gene_type == "V":
                if alignment.target_begin > 0:
                    # the input sequence should contain at least enough truncated 5' residues to reach the
                    # start of the germline V gene segment
                    if alignment.query_begin >= alignment.target_begin:
                        antibody.log("\nFORCING FULL-LENGTH V-GENE REALIGNMENT")
                        antibody.log("ORIGINAL RAW QUERY:", alignment.raw_query)
                        antibody.log("ORIGINAL RAW GERMLINE:", alignment.raw_target)
                        antibody.log(
                            "ORIGINAL ALIGNED QUERY: ", alignment.aligned_query
                        )
                        antibody.log(
                            "ORIGINAL ALIGNED GERMLINE:", alignment.aligned_target
                        )
                        antibody.log("ORIGINAL QUERY START: ", alignment.query_begin)
                        antibody.log("ORIGINAL QUERY END: ", alignment.query_end)
                        antibody.log("ORIGINAL GERMLINE START:", alignment.target_begin)
                        antibody.log("ORIGINAL GERMLINE END: ", alignment.target_end)
                        antibody.log(
                            "NUMBER OF TRUNCATED RESIDUES:", alignment.target_begin
                        )

                        trunc_length = alignment.target_begin
                        # get the truncated portion of the query sequence
                        # and prepend to the aligned query sequence
                        query_truncation = alignment.raw_query[
                            alignment.query_begin - trunc_length : alignment.query_begin
                        ]
                        new_aligned_query = query_truncation + alignment.aligned_query
                        alignment.aligned_query = new_aligned_query
                        # update the query start position
                        q_begin = alignment.query_begin - trunc_length
                        alignment.query_begin = q_begin
                        # get the truncated portion of the target (germline) sequence
                        # and prepend to the aligned target sequence
                        target_truncation = alignment.raw_target[
                            alignment.target_begin
                            - trunc_length : alignment.target_begin
                        ]
                        new_aligned_target = (
                            target_truncation + alignment.aligned_target
                        )
                        alignment.aligned_target = new_aligned_target
                        # update the target (germline) start position
                        t_begin = 0
                        alignment.target_begin = t_begin

                        antibody.log("NEW ALIGNED QUERY: ", alignment.aligned_query)
                        antibody.log("NEW ALIGNED GERMLINE:", alignment.aligned_target)

            if self.gene_type == "J":
                if alignment.target_end + 1 < len(alignment.raw_target):
                    # the input sequence should contain at least enough truncated 3' residues to reach the
                    # end of the germline J gene segment
                    query_truncation_length = len(alignment.raw_query) - (
                        alignment.query_end + 1
                    )
                    target_truncation_length = len(alignment.raw_target) - (
                        alignment.target_end + 1
                    )
                    if query_truncation_length >= target_truncation_length:
                        antibody.log("\nFORCING FULL-LENGTH V-GENE REALIGNMENT")
                        antibody.log("ORIGINAL RAW QUERY:", alignment.raw_query)
                        antibody.log("ORIGINAL RAW GERMLINE:", alignment.raw_target)
                        antibody.log(
                            "ORIGINAL ALIGNED QUERY: ", alignment.aligned_query
                        )
                        antibody.log(
                            "ORIGINAL ALIGNED GERMLINE:", alignment.aligned_target
                        )
                        antibody.log("ORIGINAL QUERY START: ", alignment.query_begin)
                        antibody.log("ORIGINAL QUERY END: ", alignment.query_end)
                        antibody.log("ORIGINAL GERMLINE START:", alignment.target_begin)
                        antibody.log("ORIGINAL GERMLINE END: ", alignment.target_end)
                        antibody.log(
                            "NUMBER OF TRUNCATED RESIDUES:", target_truncation_length
                        )

                        # get the truncated portion of the query sequence
                        # and append to the aligned query sequence
                        query_truncation = alignment.raw_query[
                            alignment.query_end
                            + 1 : alignment.query_end
                            + target_truncation_length
                            + 1
                        ]
                        new_aligned_query = alignment.aligned_query + query_truncation
                        alignment.aligned_query = new_aligned_query
                        # update the query end position
                        q_end = alignment.query_end + target_truncation_length
                        alignment.query_end = q_end
                        # get the truncated portion of the target (germline) sequence
                        # and append to the aligned target sequence
                        target_truncation = alignment.raw_target[
                            alignment.target_end
                            + 1 : alignment.target_end
                            + target_truncation_length
                            + 1
                        ]
                        new_aligned_target = (
                            alignment.aligned_target + target_truncation
                        )
                        alignment.aligned_target = new_aligned_target
                        # update the target (germline) end position
                        t_end = alignment.target_end + target_truncation_length
                        alignment.target_end = t_end

                        antibody.log("NEW ALIGNED QUERY: ", alignment.aligned_query)
                        antibody.log("NEW ALIGNED GERMLINE:", alignment.aligned_target)
        if alignment:
            self._process_realignment(antibody, alignment, query_start)
        else:
            antibody.log("GERMLINE REALIGNMENT ERROR")
            antibody.log("REALIGNMENT QUERY SEQUENCE:", query)
            antibody.log("QUERY START:", query_start)
            antibody.log("QUERY END:", query_end)

    def gapped_imgt_realignment(self):
        """
        Aligns to gapped IMGT germline sequence. Used to determine
        IMGT-formatted position numberings so that identifying
        antibody regions is simplified.
        """
        self.imgt_germline = get_imgt_germlines(
            self.db_name,
            gene_type=self.gene_type,
            receptor=self.receptor,
            gene=self.full,
        )
        query = self.germline_alignment.replace("-", "")
        aln_params = self._realignment_scoring_params(self.gene_type)
        aln_params["gap_open"] = -11
        # aln_matrix = self._get_gapped_imgt_substitution_matrix()
        self.imgt_gapped_alignment = local_alignment(
            query,
            self.imgt_germline.gapped_nt_sequence,
            # matrix=aln_matrix,
            **aln_params,
        )
        self.alignment_reading_frame = (
            (2 * (self.imgt_gapped_alignment.target_begin % 3)) % 3
        ) + (
            self.imgt_germline.coding_start - 1
        )  # IMGT coding start is 1-based
        self.coding_region = self._get_coding_region()
        self.aa_sequence = self._get_aa_sequence()
        try:
            self._imgt_numbering()
        except:
            self.exception("IMGT NUMBERING", traceback.format_exc(), sep="\n")

    def _get_gapped_imgt_substitution_matrix(self):
        residues = "ACGTN."
        m = parasail.matrix_create(residues, 3, -2)
        for i in range(len(residues)):
            m[5, i] = -3
            m[i, 5] = -3
        return m

        # matrix = {}
        # residues = ["A", "C", "G", "T", "N", "."]
        # for r1 in residues:
        #     matrix[r1] = {}
        #     for r2 in residues:
        #         if r1 == r2:
        #             score = 3
        #         elif any([r1 == ".", r2 == "."]):
        #             score = -3
        #         else:
        #             score = -2
        #         matrix[r1][r2] = score
        # return matrix

    def get_imgt_position_from_raw(self, raw):
        return self._imgt_position_from_raw.get(raw, None)

    def get_raw_position_from_imgt(self, imgt):
        return self._raw_position_from_imgt.get(imgt, None)

    def _process_realignment(self, antibody, aln, query_start):
        self.realignment = aln
        self.score = aln.score
        self.raw_query = aln.raw_query
        self.raw_germline = aln.raw_target
        self.query_alignment = aln.aligned_query
        self.germline_alignment = aln.aligned_target
        self.alignment_midline = "".join(
            [
                "|" if q == g else " "
                for q, g in zip(aln.aligned_query, aln.aligned_target)
            ]
        )
        # only update alignment start/end positions if not already annotated by aligner
        if any(
            [
                x is None
                for x in [
                    self.query_start,
                    self.query_end,
                    self.germline_start,
                    self.germline_end,
                ]
            ]
        ):
            offset = query_start if query_start is not None else 0
            self.query_start = aln.query_begin + offset
            self.query_end = aln.query_end + offset
            self.germline_start = aln.target_begin
            self.germline_end = aln.target_end
        self._fix_ambigs(antibody)
        self._find_indels(antibody)

    def _get_germline_sequence_for_realignment(self, antibody):
        """
        Identifies the appropriate germline variable gene from a database of all
        germline variable genes.

        Returns:
        --------

            str: Germline sequence, or ``None`` if the requested germline gene could not be found.
        """
        # mod_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        # db_file = os.path.join(mod_dir, 'assigners/germline_dbs/{}_{}.fasta'.format(self.species.lower(), self.gene_type))
        germ_dir = get_germline_database_directory(self.db_name, self.receptor)
        antibody.log("GERMLINE DIRECTORY: {}".format(germ_dir))
        db_file = os.path.join(
            germ_dir, "ungapped/{}.fasta".format(self.gene_type.lower())
        )
        antibody.log("GERMLINE DB FILE: {}".format(db_file))
        try:
            for s in SeqIO.parse(open(db_file), "fasta"):
                if s.id == self.raw_assignment:
                    return str(s.seq)
            # TODO: log that the germline gene wasn't found in the database file
            return None
        except:
            # TODO: log that the germline database file couldn't be found
            return None

    def _imgt_numbering(self):
        aln_start = self.query_start
        aln_pos = 0
        imgt_start = self.imgt_gapped_alignment.target_begin + 1
        imgt_pos = imgt_start
        raw_position_from_imgt = {}
        imgt_position_from_raw = {}

        # imgt_start_offset is for J-genes only. Since the first position of the gapped IMGT
        # V-gene is the first position of the antibody seqeunce, IMGT numbering of the
        # V-gene should start at position 1. Not true with J-genes.
        # When processing V-genes, imgt_start_offset will always be 0.
        imgt_start_offset = self._get_imgt_start_offset()

        # Because we're iterating over the alignment but also want to track the raw (oriented_input)
        # query position, we need to adjust the alignment numbering in case there's a deletion
        # in the query sequence. With a deletion, the alignment position increases, but the raw
        # position shouldn't. query_del_adjustment allows us to adjust the alignment position so that
        # the raw query position is accurately tracked.
        query_del_adjustment = 0

        for gl in self.imgt_germline.gapped_nt_sequence[imgt_start:]:
            # only compute IMGT AA position numbers for V-genes. J-gene numbering is dependent on the CDR3 length
            # (due to IMGT's CDR3 naming conventions), so the junction must be annotated before J-gene AA position
            # numbers can be assigned.
            # if self.gene_type == 'V':

            # check to see if the position we're looking at is the start of a codon
            is_codon_start = (imgt_pos + imgt_start_offset) % 3 == 1

            # start by adding the aa position number (only if we're at the start of a codon)
            if is_codon_start:
                codon = self.germline_alignment[aln_pos : aln_pos + 3]
                if codon.count("-") >= 2:
                    self.imgt_aa_positions.append(None)
                else:
                    self.imgt_aa_positions.append(
                        (imgt_pos + imgt_start_offset + 2) / 3
                    )

            # If the gapped IMGT germline is '.' (indicating a gap introduced by IMGT for numbering purposes),
            # we only need to increment the IMGT position and indicate the lack of sequence at the IMGT position.
            if gl == ".":
                raw_position_from_imgt[imgt_pos + imgt_start_offset] = None
                imgt_pos += 1
                continue

            # If there's a gap in the query alignment (deletion in the query sequence)
            # there's no equivalent IMGT position in the query.
            if self.query_alignment[aln_pos] == "-":
                raw_position_from_imgt[imgt_pos + imgt_start_offset] = None
                aln_pos += 1
                imgt_pos += 1
                query_del_adjustment += 1
                continue

            # If there's a gap in the germline alignment (insertion in the query sequence)
            # we need to increment the alignment position until the insertion is finished.
            # Since there isn't a germline IMGT position that's equivalent to the query position,
            # we don't want to increment the IMGT position or continue iterating through the IMGT
            # germline sequence.
            if self.germline_alignment[aln_pos] == "-":
                self.log("INFO: Found an insertion in the query sequence!")
                while self.germline_alignment[aln_pos] == "-":
                    imgt_position_from_raw[
                        aln_pos + self.query_start - query_del_adjustment
                    ] = None
                    self.imgt_nt_positions.append(None)
                    aln_pos += 1

            # if the gapped IMGT germline isn't '.' and there's not an insertion in the query
            # sequence (or we've already iterated past it), we record both the IMGT position
            # and the raw (oriented_input) position and increment both the aligned and IMGt positions.
            raw_position_from_imgt[imgt_pos + imgt_start_offset] = (
                aln_pos + aln_start - query_del_adjustment
            )
            imgt_position_from_raw[aln_pos + aln_start - query_del_adjustment] = (
                imgt_pos + imgt_start_offset
            )
            self.imgt_nt_positions.append(imgt_pos + imgt_start_offset)
            aln_pos += 1
            imgt_pos += 1
            if aln_pos >= len(self.germline_alignment):
                break
        self._raw_position_from_imgt = raw_position_from_imgt
        self._imgt_position_from_raw = imgt_position_from_raw
        if self.insertions or self.deletions:
            self._calculate_imgt_indel_positions()

    def _get_imgt_start_offset(self):
        if self.gene_type == "V":
            return 0
        # find the start of FR4 in the IMGT gapped germline gene
        end_res = "W" if self.chain == "heavy" else "F"
        for i, res in enumerate(self.imgt_germline.ungapped_aa_sequence):
            if (
                res == end_res
                and end_res not in self.imgt_germline.ungapped_aa_sequence[i + 1 :]
            ):
                nts_from_start_to_fr4 = (self.imgt_germline.coding_start) + (i * 3)
                break
        # the IMGT start offset is the conserved FR4 start position (352)
        # minus the number of nts from the start of the germline gene to FR4
        return 352 - nts_from_start_to_fr4

    def _get_coding_region(self):
        coding_region = self.query_alignment[self.alignment_reading_frame :].replace(
            "-", ""
        )
        truncation = len(coding_region) % 3
        if truncation > 0:
            coding_region = coding_region[:-truncation]
        return coding_region

    def _get_aa_sequence(self):
        return Seq(self.coding_region).translate()

    def _fix_ambigs(self, antibody):
        """
        Fixes ambiguous nucleotides by replacing them with the germline nucleotide.
        """
        self.query_alignment = "".join(
            [
                q if q.upper() != "N" else g
                for q, g in zip(self.query_alignment, self.germline_alignment)
            ]
        )
        # don't forget to also correct ambigs in the oriented_input sequence
        antibody.oriented_input.sequence = (
            antibody.oriented_input.sequence[: self.query_start]
            + self.query_alignment.replace("-", "")
            + antibody.oriented_input.sequence[self.query_end + 1 :]
        )

    def _indel_check(self):
        if any(["-" in self.query_alignment, "-" in self.germline_alignment]):
            return True
        return False

    def _find_indels(self, antibody):
        """
        Identifies and annotates indels in the query sequence.
        """
        if self._indel_check():
            self.insertions = indels.find_insertions(antibody, self)
            if self.insertions:
                # only set self.has_insertion to 'yes' if the sequence has an in-frame insertion
                if "yes" in [ins["in frame"] for ins in self.insertions]:
                    self.has_insertion = "yes"
            self.deletions = indels.find_deletions(antibody, self)
            if self.deletions:
                # only set self.has_deletion to 'yes' if the sequence has an in-frame deletion
                if "yes" in [deletion["in frame"] for deletion in self.deletions]:
                    self.has_deletion = "yes"

    def _calculate_imgt_indel_positions(self):
        if self.insertions:
            for i in self.insertions:
                # since there's possibly not a direct IMGT correlate to the position
                # at the start of the insertion, need to find the closest IMGT correlate
                raw_pos = i.raw_position
                while self.get_imgt_position_from_raw(raw_pos) is None:
                    raw_pos -= 1
                i.imgt_position = self.get_imgt_position_from_raw(raw_pos)
                i.imgt_codon = int(math.ceil(i.imgt_position / 3.0))
        if self.deletions:
            for d in self.deletions:
                # since there's possibly not a direct IMGT correlate to the position
                # at the start of the deletion, need to find the closest IMGT correlate
                raw_pos = d.raw_position
                while self.get_imgt_position_from_raw(raw_pos) is None:
                    raw_pos -= 1
                d.imgt_position = self.get_imgt_position_from_raw(raw_pos)
                d.imgt_codon = int(math.ceil(d.imgt_position / 3.0))

    @staticmethod
    def _realignment_scoring_params(gene):
        """
        Returns realignment scoring paramaters for a given gene type.

        Args:

            gene (str): the gene type ('V', 'D', or 'J')


        Returns:

            dict: realignment scoring parameters
        """
        scores = {
            "V": {"match": 3, "mismatch": -2, "gap_open": -22, "gap_extend": -1},
            "D": {"match": 3, "mismatch": -2, "gap_open": -22, "gap_extend": -1},
            "J": {"match": 3, "mismatch": -2, "gap_open": -22, "gap_extend": -1},
        }
        return scores[gene]


def get_germline_database_directory(species, receptor="bcr"):
    species = species.lower()
    receptor = receptor.lower()
    addon_dir = os.path.expanduser(f"~/.abstar/germline_dbs/{receptor}")
    if os.path.isdir(addon_dir):
        if species.lower() in [os.path.basename(d[0]) for d in os.walk(addon_dir)]:
            return os.path.join(addon_dir, species.lower())
    mod_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    return os.path.join(mod_dir, f"assigners/germline_dbs/{receptor}/{species}")


def get_imgt_germlines(db_name, gene_type, receptor="bcr", gene=None):
    """
    Returns one or more IMGTGermlineGene objects that each contain a single IMGT-gapped germline gene.

    Args:
    -----

        species (str): Species for which the germline genes should be obtained.

        gene_type (str): Options are 'V', 'D', and 'J'.

        receptor (str): Options are ``'bcr'`` and ``'tcr'``.

        gene (str): Full name of a germline gene (using IMGT-style names, like IGHV1-2*02).
                    If provided, a single ``IMGTGermlineGene`` object will be returned, or None if the
                    specified gene could not be found. If not provided, a list of ``IMGTGermlineGene``
                    objects for all germline genes matching the ``species`` and ``gene_type`` will be returned.

    Returns:
    --------

        IMGTGermlineGene: a single ``IMGTGermlineGene`` object (if ``gene`` is provided) or a list of
                          ``IMGTGermlineGene`` objects. If no sequences match the criteria or if a germline
                          database for the requested species is not found, ``None`` is returned.
    """
    # mod_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    # db_file = os.path.join(mod_dir, 'assigners/germline_dbs/imgt_gapped/{}_{}_imgt-gapped.fasta'.format(species, gene_type))
    germ_dir = get_germline_database_directory(db_name, receptor.lower())
    db_file = os.path.join(germ_dir, "imgt_gapped/{}.fasta".format(gene_type.lower()))
    try:
        germs = [IMGTGermlineGene(g) for g in SeqIO.parse(open(db_file, "r"), "fasta")]
    except:
        # TODO: log that the germline database file couldn't be found

        # print('Could not locate the IMGT germline database file ({}).'.format(db_file))
        # print(traceback.format_exc())

        return None
    if gene is None:
        return germs
    try:
        return [g for g in germs if g.name == gene][0]
    except IndexError:
        # print('Could not locate the IMGT germline gene ({}).'.format(gene))
        # print(traceback.format_exc())

        return None


def get_germlines(db_name, gene_type, receptor="bcr", chain=None, gene=None):
    """
    Returns one or more IMGTGermlineGene objects that each contain a single IMGT-gapped germline gene.

    Args:
    -----

        species (str): Species for which the germline genes should be obtained.

        gene_type (str): Options are ``'V'``, ``'D'``, and ``'J'``.

        receptor (str): Options are ``'bcr'`` and ``'tcr'``.

        chain (str): Options are ``'heavy'``, ``'kappa'``, and ``'lambda'`` for BCRs and ``'alpha'``,
        ``'beta'``, ``'gamma'``, and ``'delta'`` for TCRs. If not provided, germline sequences from
        all chains are returned.

        gene (str): Full name of a germline gene (using IMGT-style names, like ``'IGHV1-2*02'``).
        If provided, a single ``IMGTGermlineGene`` object will be returned, or ``None`` if the
        specified gene could not be found. If not provided, a list of ``IMGTGermlineGene`` objects
        for all germline genes matching the ``species``, ``receptor`` and ``gene_type`` will be returned.

    Returns:
    --------

        IMGTGermlineGene: a single ``IMGTGermlineGene`` object (if ``gene`` is provided) or a list of
                          ``IMGTGermlineGene`` objects. If no sequences match the criteria or if a germline
                          database for the requested species is not found, ``None`` is returned.
    """
    germs = get_imgt_germlines(db_name, gene_type, receptor=receptor.lower(), gene=gene)
    if germs is None:
        return germs
    if chain is not None:
        c = chain[0].upper()
        germs = [g for g in germs if g.name[2] == c]
    if len(germs) == 1:
        return germs[0]
    return germs


class IMGTGermlineGene(object):
    """docstring for IMGTGermlineGene"""

    # species_lookup exists to translate IMGT names to AbStar germline DB names.
    # if the IMGT species isn't in species_lookup, AbStar will try to use the IMGT
    # species name directly.
    species_lookup = {"homo sapiens": "human"}

    def __init__(self, sequence, species=None):
        self.raw_sequence = Sequence(str(sequence.seq), id=sequence.description)
        self._species = species
        self.gapped_nt_sequence = self.raw_sequence.sequence
        self.ungapped_nt_sequence = self.gapped_nt_sequence.replace(".", "")

    @lazy_property
    def accession(self):
        return self.raw_sequence.id.split("|")[0].strip()

    @lazy_property
    def name(self):
        return self.raw_sequence.id.split("|")[1].strip()

    @property
    def species(self):
        if self._species is None:
            species = self.raw_sequence.id.split("|")[2].strip().lower()
            self._species = self.species_lookup.get(species, species)
        return self._species

    @lazy_property
    def functionality(self):
        return self.raw_sequence.id.split("|")[3].strip()

    @lazy_property
    def gene_type(self):
        return self.raw_sequence.id.split("|")[4].strip()[0].upper()

    @lazy_property
    def coding_start(self):
        # NOTE: uses 1-based indexing, so need to adjust if using for slicing
        return int(self.raw_sequence.id.split("|")[7].strip())

    @lazy_property
    def nt_length(self):
        return int(self.raw_sequence.id.split("|")[12].strip().split("+")[0])

    @lazy_property
    def gap_length(self):
        return int(
            self.raw_sequence.id.split("|")[12].strip().split("+")[1].split("=")[0]
        )

    @lazy_property
    def total_length(self):
        return int(self.raw_sequence.id.split("|")[12].strip().split("=")[1])

    @lazy_property
    def partial(self):
        p = []
        partial = self.raw_sequence.id.split("|")[13]
        if "3'" in partial:
            p.append("3'")
        if "5'" in partial:
            p.append("5'")
        return p

    @lazy_property
    def is_rev_comp(self):
        if self.raw_sequence.id.split("|")[14].strip() == "rev-compl":
            is_rev_comp = True
        else:
            is_rev_comp = False
        return is_rev_comp

    @lazy_property
    def gapped_aa_sequence(self):
        res = []
        coding = self.gapped_nt_sequence[self.coding_start - 1 :]
        for codon in (coding[pos : pos + 3] for pos in range(0, len(coding), 3)):
            if len(codon) != 3:
                continue
            if codon == "...":
                res.append(".")
            else:
                res.append(codon_lookup.get(codon, "X"))
        return "".join(res)

    @lazy_property
    def ungapped_aa_sequence(self):
        return self.gapped_aa_sequence.replace(".", "")
